/*
 * Copyright (c) 2012 Taobao.
 * Written by Tao Ma <boyu.mt@taobao.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "ldiskfs_jbd2.h"
#include "ldiskfs.h"
#include "xattr.h"
#include "truncate.h"
#include <linux/fiemap.h>

#define LDISKFS_XATTR_SYSTEM_DATA	"data"
#define LDISKFS_MIN_INLINE_DATA_SIZE	((sizeof(__le32) * LDISKFS_N_BLOCKS))
#define LDISKFS_INLINE_DOTDOT_OFFSET	2
#define LDISKFS_INLINE_DOTDOT_SIZE		4

static int ldiskfs_get_inline_size(struct inode *inode)
{
	if (LDISKFS_I(inode)->i_inline_off)
		return LDISKFS_I(inode)->i_inline_size;

	return 0;
}

static int get_max_inline_xattr_value_size(struct inode *inode,
					   struct ldiskfs_iloc *iloc)
{
	struct ldiskfs_xattr_ibody_header *header;
	struct ldiskfs_xattr_entry *entry;
	struct ldiskfs_inode *raw_inode;
	int free, min_offs;

	min_offs = LDISKFS_SB(inode->i_sb)->s_inode_size -
			LDISKFS_GOOD_OLD_INODE_SIZE -
			LDISKFS_I(inode)->i_extra_isize -
			sizeof(struct ldiskfs_xattr_ibody_header);

	/*
	 * We need to subtract another sizeof(__u32) since an in-inode xattr
	 * needs an empty 4 bytes to indicate the gap between the xattr entry
	 * and the name/value pair.
	 */
	if (!ldiskfs_test_inode_state(inode, LDISKFS_STATE_XATTR))
		return LDISKFS_XATTR_SIZE(min_offs -
			LDISKFS_XATTR_LEN(strlen(LDISKFS_XATTR_SYSTEM_DATA)) -
			LDISKFS_XATTR_ROUND - sizeof(__u32));

	raw_inode = ldiskfs_raw_inode(iloc);
	header = IHDR(inode, raw_inode);
	entry = IFIRST(header);

	/* Compute min_offs. */
	for (; !IS_LAST_ENTRY(entry); entry = LDISKFS_XATTR_NEXT(entry)) {
		if (!entry->e_value_inum && entry->e_value_size) {
			size_t offs = le16_to_cpu(entry->e_value_offs);
			if (offs < min_offs)
				min_offs = offs;
		}
	}
	free = min_offs -
		((void *)entry - (void *)IFIRST(header)) - sizeof(__u32);

	if (LDISKFS_I(inode)->i_inline_off) {
		entry = (struct ldiskfs_xattr_entry *)
			((void *)raw_inode + LDISKFS_I(inode)->i_inline_off);

		free += LDISKFS_XATTR_SIZE(le32_to_cpu(entry->e_value_size));
		goto out;
	}

	free -= LDISKFS_XATTR_LEN(strlen(LDISKFS_XATTR_SYSTEM_DATA));

	if (free > LDISKFS_XATTR_ROUND)
		free = LDISKFS_XATTR_SIZE(free - LDISKFS_XATTR_ROUND);
	else
		free = 0;

out:
	return free;
}

/*
 * Get the maximum size we now can store in an inode.
 * If we can't find the space for a xattr entry, don't use the space
 * of the extents since we have no space to indicate the inline data.
 */
int ldiskfs_get_max_inline_size(struct inode *inode)
{
	int error, max_inline_size;
	struct ldiskfs_iloc iloc;

	if (LDISKFS_I(inode)->i_extra_isize == 0)
		return 0;

	error = ldiskfs_get_inode_loc(inode, &iloc);
	if (error) {
		ldiskfs_error_inode(inode, __func__, __LINE__, 0,
				 "can't get inode location %lu",
				 inode->i_ino);
		return 0;
	}

	down_read(&LDISKFS_I(inode)->xattr_sem);
	max_inline_size = get_max_inline_xattr_value_size(inode, &iloc);
	up_read(&LDISKFS_I(inode)->xattr_sem);

	brelse(iloc.bh);

	if (!max_inline_size)
		return 0;

	return max_inline_size + LDISKFS_MIN_INLINE_DATA_SIZE;
}

/*
 * this function does not take xattr_sem, which is OK because it is
 * currently only used in a code path coming form ldiskfs_iget, before
 * the new inode has been unlocked
 */
int ldiskfs_find_inline_data_nolock(struct inode *inode)
{
	struct ldiskfs_xattr_ibody_find is = {
		.s = { .not_found = -ENODATA, },
	};
	struct ldiskfs_xattr_info i = {
		.name_index = LDISKFS_XATTR_INDEX_SYSTEM,
		.name = LDISKFS_XATTR_SYSTEM_DATA,
	};
	int error;

	if (LDISKFS_I(inode)->i_extra_isize == 0)
		return 0;

	error = ldiskfs_get_inode_loc(inode, &is.iloc);
	if (error)
		return error;

	error = ldiskfs_xattr_ibody_find(inode, &i, &is);
	if (error)
		goto out;

	if (!is.s.not_found) {
		LDISKFS_I(inode)->i_inline_off = (u16)((void *)is.s.here -
					(void *)ldiskfs_raw_inode(&is.iloc));
		LDISKFS_I(inode)->i_inline_size = LDISKFS_MIN_INLINE_DATA_SIZE +
				le32_to_cpu(is.s.here->e_value_size);
		ldiskfs_set_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
	}
out:
	brelse(is.iloc.bh);
	return error;
}

static int ldiskfs_read_inline_data(struct inode *inode, void *buffer,
				 unsigned int len,
				 struct ldiskfs_iloc *iloc)
{
	struct ldiskfs_xattr_entry *entry;
	struct ldiskfs_xattr_ibody_header *header;
	int cp_len = 0;
	struct ldiskfs_inode *raw_inode;

	if (!len)
		return 0;

	BUG_ON(len > LDISKFS_I(inode)->i_inline_size);

	cp_len = len < LDISKFS_MIN_INLINE_DATA_SIZE ?
			len : LDISKFS_MIN_INLINE_DATA_SIZE;

	raw_inode = ldiskfs_raw_inode(iloc);
	memcpy(buffer, (void *)(raw_inode->i_block), cp_len);

	len -= cp_len;
	buffer += cp_len;

	if (!len)
		goto out;

	header = IHDR(inode, raw_inode);
	entry = (struct ldiskfs_xattr_entry *)((void *)raw_inode +
					    LDISKFS_I(inode)->i_inline_off);
	len = min_t(unsigned int, len,
		    (unsigned int)le32_to_cpu(entry->e_value_size));

	memcpy(buffer,
	       (void *)IFIRST(header) + le16_to_cpu(entry->e_value_offs), len);
	cp_len += len;

out:
	return cp_len;
}

/*
 * write the buffer to the inline inode.
 * If 'create' is set, we don't need to do the extra copy in the xattr
 * value since it is already handled by ldiskfs_xattr_ibody_inline_set.
 * That saves us one memcpy.
 */
static void ldiskfs_write_inline_data(struct inode *inode, struct ldiskfs_iloc *iloc,
				   void *buffer, loff_t pos, unsigned int len)
{
	struct ldiskfs_xattr_entry *entry;
	struct ldiskfs_xattr_ibody_header *header;
	struct ldiskfs_inode *raw_inode;
	int cp_len = 0;

	BUG_ON(!LDISKFS_I(inode)->i_inline_off);
	BUG_ON(pos + len > LDISKFS_I(inode)->i_inline_size);

	raw_inode = ldiskfs_raw_inode(iloc);
	buffer += pos;

	if (pos < LDISKFS_MIN_INLINE_DATA_SIZE) {
		cp_len = pos + len > LDISKFS_MIN_INLINE_DATA_SIZE ?
			 LDISKFS_MIN_INLINE_DATA_SIZE - pos : len;
		memcpy((void *)raw_inode->i_block + pos, buffer, cp_len);

		len -= cp_len;
		buffer += cp_len;
		pos += cp_len;
	}

	if (!len)
		return;

	pos -= LDISKFS_MIN_INLINE_DATA_SIZE;
	header = IHDR(inode, raw_inode);
	entry = (struct ldiskfs_xattr_entry *)((void *)raw_inode +
					    LDISKFS_I(inode)->i_inline_off);

	memcpy((void *)IFIRST(header) + le16_to_cpu(entry->e_value_offs) + pos,
	       buffer, len);
}

static int ldiskfs_create_inline_data(handle_t *handle,
				   struct inode *inode, unsigned len)
{
	int error;
	void *value = NULL;
	struct ldiskfs_xattr_ibody_find is = {
		.s = { .not_found = -ENODATA, },
	};
	struct ldiskfs_xattr_info i = {
		.name_index = LDISKFS_XATTR_INDEX_SYSTEM,
		.name = LDISKFS_XATTR_SYSTEM_DATA,
	};

	error = ldiskfs_get_inode_loc(inode, &is.iloc);
	if (error)
		return error;

	BUFFER_TRACE(is.iloc.bh, "get_write_access");
	error = ldiskfs_journal_get_write_access(handle, is.iloc.bh);
	if (error)
		goto out;

	if (len > LDISKFS_MIN_INLINE_DATA_SIZE) {
		value = LDISKFS_ZERO_XATTR_VALUE;
		len -= LDISKFS_MIN_INLINE_DATA_SIZE;
	} else {
		value = "";
		len = 0;
	}

	/* Insert the the xttr entry. */
	i.value = value;
	i.value_len = len;

	error = ldiskfs_xattr_ibody_find(inode, &i, &is);
	if (error)
		goto out;

	BUG_ON(!is.s.not_found);

	error = ldiskfs_xattr_ibody_inline_set(handle, inode, &i, &is);
	if (error) {
		if (error == -ENOSPC)
			ldiskfs_clear_inode_state(inode,
					       LDISKFS_STATE_MAY_INLINE_DATA);
		goto out;
	}

	memset((void *)ldiskfs_raw_inode(&is.iloc)->i_block,
		0, LDISKFS_MIN_INLINE_DATA_SIZE);

	LDISKFS_I(inode)->i_inline_off = (u16)((void *)is.s.here -
				      (void *)ldiskfs_raw_inode(&is.iloc));
	LDISKFS_I(inode)->i_inline_size = len + LDISKFS_MIN_INLINE_DATA_SIZE;
	ldiskfs_clear_inode_flag(inode, LDISKFS_INODE_EXTENTS);
	ldiskfs_set_inode_flag(inode, LDISKFS_INODE_INLINE_DATA);
	get_bh(is.iloc.bh);
	error = ldiskfs_mark_iloc_dirty(handle, inode, &is.iloc);

out:
	brelse(is.iloc.bh);
	return error;
}

static int ldiskfs_update_inline_data(handle_t *handle, struct inode *inode,
				   unsigned int len)
{
	int error;
	void *value = NULL;
	struct ldiskfs_xattr_ibody_find is = {
		.s = { .not_found = -ENODATA, },
	};
	struct ldiskfs_xattr_info i = {
		.name_index = LDISKFS_XATTR_INDEX_SYSTEM,
		.name = LDISKFS_XATTR_SYSTEM_DATA,
	};

	/* If the old space is ok, write the data directly. */
	if (len <= LDISKFS_I(inode)->i_inline_size)
		return 0;

	error = ldiskfs_get_inode_loc(inode, &is.iloc);
	if (error)
		return error;

	error = ldiskfs_xattr_ibody_find(inode, &i, &is);
	if (error)
		goto out;

	BUG_ON(is.s.not_found);

	len -= LDISKFS_MIN_INLINE_DATA_SIZE;
	value = kzalloc(len, GFP_NOFS);
	if (!value) {
		error = -ENOMEM;
		goto out;
	}

	error = ldiskfs_xattr_ibody_get(inode, i.name_index, i.name,
				     value, len);
	if (error == -ENODATA)
		goto out;

	BUFFER_TRACE(is.iloc.bh, "get_write_access");
	error = ldiskfs_journal_get_write_access(handle, is.iloc.bh);
	if (error)
		goto out;

	/* Update the xttr entry. */
	i.value = value;
	i.value_len = len;

	error = ldiskfs_xattr_ibody_inline_set(handle, inode, &i, &is);
	if (error)
		goto out;

	LDISKFS_I(inode)->i_inline_off = (u16)((void *)is.s.here -
				      (void *)ldiskfs_raw_inode(&is.iloc));
	LDISKFS_I(inode)->i_inline_size = LDISKFS_MIN_INLINE_DATA_SIZE +
				le32_to_cpu(is.s.here->e_value_size);
	ldiskfs_set_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
	get_bh(is.iloc.bh);
	error = ldiskfs_mark_iloc_dirty(handle, inode, &is.iloc);

out:
	kfree(value);
	brelse(is.iloc.bh);
	return error;
}

static int ldiskfs_prepare_inline_data(handle_t *handle, struct inode *inode,
				    unsigned int len)
{
	int ret, size;
	struct ldiskfs_inode_info *ei = LDISKFS_I(inode);

	if (!ldiskfs_test_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA))
		return -ENOSPC;

	size = ldiskfs_get_max_inline_size(inode);
	if (size < len)
		return -ENOSPC;

	down_write(&LDISKFS_I(inode)->xattr_sem);

	if (ei->i_inline_off)
		ret = ldiskfs_update_inline_data(handle, inode, len);
	else
		ret = ldiskfs_create_inline_data(handle, inode, len);

	up_write(&LDISKFS_I(inode)->xattr_sem);

	return ret;
}

static int ldiskfs_destroy_inline_data_nolock(handle_t *handle,
					   struct inode *inode)
{
	struct ldiskfs_inode_info *ei = LDISKFS_I(inode);
	struct ldiskfs_xattr_ibody_find is = {
		.s = { .not_found = 0, },
	};
	struct ldiskfs_xattr_info i = {
		.name_index = LDISKFS_XATTR_INDEX_SYSTEM,
		.name = LDISKFS_XATTR_SYSTEM_DATA,
		.value = NULL,
		.value_len = 0,
	};
	int error;

	if (!ei->i_inline_off)
		return 0;

	error = ldiskfs_get_inode_loc(inode, &is.iloc);
	if (error)
		return error;

	error = ldiskfs_xattr_ibody_find(inode, &i, &is);
	if (error)
		goto out;

	BUFFER_TRACE(is.iloc.bh, "get_write_access");
	error = ldiskfs_journal_get_write_access(handle, is.iloc.bh);
	if (error)
		goto out;

	error = ldiskfs_xattr_ibody_inline_set(handle, inode, &i, &is);
	if (error)
		goto out;

	memset((void *)ldiskfs_raw_inode(&is.iloc)->i_block,
		0, LDISKFS_MIN_INLINE_DATA_SIZE);
	memset(ei->i_data, 0, LDISKFS_MIN_INLINE_DATA_SIZE);

	if (LDISKFS_HAS_INCOMPAT_FEATURE(inode->i_sb,
				      LDISKFS_FEATURE_INCOMPAT_EXTENTS)) {
		if (S_ISDIR(inode->i_mode) ||
		    S_ISREG(inode->i_mode) || S_ISLNK(inode->i_mode)) {
			ldiskfs_set_inode_flag(inode, LDISKFS_INODE_EXTENTS);
			ldiskfs_ext_tree_init(handle, inode);
		}
	}
	ldiskfs_clear_inode_flag(inode, LDISKFS_INODE_INLINE_DATA);

	get_bh(is.iloc.bh);
	error = ldiskfs_mark_iloc_dirty(handle, inode, &is.iloc);

	LDISKFS_I(inode)->i_inline_off = 0;
	LDISKFS_I(inode)->i_inline_size = 0;
	ldiskfs_clear_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
out:
	brelse(is.iloc.bh);
	if (error == -ENODATA)
		error = 0;
	return error;
}

static int ldiskfs_read_inline_page(struct inode *inode, struct page *page)
{
	void *kaddr;
	int ret = 0;
	size_t len;
	struct ldiskfs_iloc iloc;

	BUG_ON(!PageLocked(page));
	BUG_ON(!ldiskfs_has_inline_data(inode));
	BUG_ON(page->index);

	if (!LDISKFS_I(inode)->i_inline_off) {
		ldiskfs_warning(inode->i_sb, "inode %lu doesn't have inline data.",
			     inode->i_ino);
		goto out;
	}

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		goto out;

	len = min_t(size_t, ldiskfs_get_inline_size(inode), i_size_read(inode));
	kaddr = kmap_atomic(page);
	ret = ldiskfs_read_inline_data(inode, kaddr, len, &iloc);
	flush_dcache_page(page);
	kunmap_atomic(kaddr);
	zero_user_segment(page, len, PAGE_CACHE_SIZE);
	SetPageUptodate(page);
	brelse(iloc.bh);

out:
	return ret;
}

int ldiskfs_readpage_inline(struct inode *inode, struct page *page)
{
	int ret = 0;

	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		up_read(&LDISKFS_I(inode)->xattr_sem);
		return -EAGAIN;
	}

	/*
	 * Current inline data can only exist in the 1st page,
	 * So for all the other pages, just set them uptodate.
	 */
	if (!page->index)
		ret = ldiskfs_read_inline_page(inode, page);
	else if (!PageUptodate(page)) {
		zero_user_segment(page, 0, PAGE_CACHE_SIZE);
		SetPageUptodate(page);
	}

	up_read(&LDISKFS_I(inode)->xattr_sem);

	unlock_page(page);
	return ret >= 0 ? 0 : ret;
}

static int ldiskfs_convert_inline_data_to_extent(struct address_space *mapping,
					      struct inode *inode,
					      unsigned flags)
{
	int ret, needed_blocks;
	handle_t *handle = NULL;
	int retries = 0, sem_held = 0;
	struct page *page = NULL;
	unsigned from, to;
	struct ldiskfs_iloc iloc;

	if (!ldiskfs_has_inline_data(inode)) {
		/*
		 * clear the flag so that no new write
		 * will trap here again.
		 */
		ldiskfs_clear_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
		return 0;
	}

	needed_blocks = ldiskfs_writepage_trans_blocks(inode);

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		return ret;

retry:
	handle = ldiskfs_journal_start(inode, LDISKFS_HT_WRITE_PAGE, needed_blocks);
	if (IS_ERR(handle)) {
		ret = PTR_ERR(handle);
		handle = NULL;
		goto out;
	}

	/* We cannot recurse into the filesystem as the transaction is already
	 * started */
	flags |= AOP_FLAG_NOFS;

	page = grab_cache_page_write_begin(mapping, 0, flags);
	if (!page) {
		ret = -ENOMEM;
		goto out;
	}

	down_write(&LDISKFS_I(inode)->xattr_sem);
	sem_held = 1;
	/* If some one has already done this for us, just exit. */
	if (!ldiskfs_has_inline_data(inode)) {
		ret = 0;
		goto out;
	}

	from = 0;
	to = ldiskfs_get_inline_size(inode);
	if (!PageUptodate(page)) {
		ret = ldiskfs_read_inline_page(inode, page);
		if (ret < 0)
			goto out;
	}

	ret = ldiskfs_destroy_inline_data_nolock(handle, inode);
	if (ret)
		goto out;

	if (ldiskfs_should_dioread_nolock(inode))
		ret = __block_write_begin(page, from, to, ldiskfs_get_block_write);
	else
		ret = __block_write_begin(page, from, to, ldiskfs_get_block);

	if (!ret && ldiskfs_should_journal_data(inode)) {
		ret = ldiskfs_walk_page_buffers(handle, page_buffers(page),
					     from, to, NULL,
					     do_journal_get_write_access);
	}

	if (ret) {
		unlock_page(page);
		page_cache_release(page);
		page = NULL;
		ldiskfs_orphan_add(handle, inode);
		up_write(&LDISKFS_I(inode)->xattr_sem);
		sem_held = 0;
		ldiskfs_journal_stop(handle);
		handle = NULL;
		ldiskfs_truncate_failed_write(inode);
		/*
		 * If truncate failed early the inode might
		 * still be on the orphan list; we need to
		 * make sure the inode is removed from the
		 * orphan list in that case.
		 */
		if (inode->i_nlink)
			ldiskfs_orphan_del(NULL, inode);
	}

	if (ret == -ENOSPC && ldiskfs_should_retry_alloc(inode->i_sb, &retries))
		goto retry;

	if (page)
		block_commit_write(page, from, to);
out:
	if (page) {
		unlock_page(page);
		page_cache_release(page);
	}
	if (sem_held)
		up_write(&LDISKFS_I(inode)->xattr_sem);
	if (handle)
		ldiskfs_journal_stop(handle);
	brelse(iloc.bh);
	return ret;
}

/*
 * Try to write data in the inode.
 * If the inode has inline data, check whether the new write can be
 * in the inode also. If not, create the page the handle, move the data
 * to the page make it update and let the later codes create extent for it.
 */
int ldiskfs_try_to_write_inline_data(struct address_space *mapping,
				  struct inode *inode,
				  loff_t pos, unsigned len,
				  unsigned flags,
				  struct page **pagep)
{
	int ret;
	handle_t *handle;
	struct page *page;
	struct ldiskfs_iloc iloc;

	if (pos + len > ldiskfs_get_max_inline_size(inode))
		goto convert;

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		return ret;

	/*
	 * The possible write could happen in the inode,
	 * so try to reserve the space in inode first.
	 */
	handle = ldiskfs_journal_start(inode, LDISKFS_HT_INODE, 1);
	if (IS_ERR(handle)) {
		ret = PTR_ERR(handle);
		handle = NULL;
		goto out;
	}

	ret = ldiskfs_prepare_inline_data(handle, inode, pos + len);
	if (ret && ret != -ENOSPC)
		goto out;

	/* We don't have space in inline inode, so convert it to extent. */
	if (ret == -ENOSPC) {
		ldiskfs_journal_stop(handle);
		brelse(iloc.bh);
		goto convert;
	}

	flags |= AOP_FLAG_NOFS;

	page = grab_cache_page_write_begin(mapping, 0, flags);
	if (!page) {
		ret = -ENOMEM;
		goto out;
	}

	*pagep = page;
	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		ret = 0;
		unlock_page(page);
		page_cache_release(page);
		goto out_up_read;
	}

	if (!PageUptodate(page)) {
		ret = ldiskfs_read_inline_page(inode, page);
		if (ret < 0) {
			unlock_page(page);
			put_page(page);
			goto out_up_read;
		}
	}

	ret = 1;
	handle = NULL;
out_up_read:
	up_read(&LDISKFS_I(inode)->xattr_sem);
out:
	if (handle)
		ldiskfs_journal_stop(handle);
	brelse(iloc.bh);
	return ret;
convert:
	return ldiskfs_convert_inline_data_to_extent(mapping,
						  inode, flags);
}

int ldiskfs_write_inline_data_end(struct inode *inode, loff_t pos, unsigned len,
			       unsigned copied, struct page *page)
{
	int ret;
	void *kaddr;
	struct ldiskfs_iloc iloc;

	if (unlikely(copied < len)) {
		if (!PageUptodate(page)) {
			copied = 0;
			goto out;
		}
	}

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret) {
		ldiskfs_std_error(inode->i_sb, ret);
		copied = 0;
		goto out;
	}

	down_write(&LDISKFS_I(inode)->xattr_sem);
	BUG_ON(!ldiskfs_has_inline_data(inode));

	kaddr = kmap_atomic(page);
	ldiskfs_write_inline_data(inode, &iloc, kaddr, pos, len);
	kunmap_atomic(kaddr);
	SetPageUptodate(page);
	/* clear page dirty so that writepages wouldn't work for us. */
	ClearPageDirty(page);

	up_write(&LDISKFS_I(inode)->xattr_sem);
	brelse(iloc.bh);
out:
	return copied;
}

struct buffer_head *
ldiskfs_journalled_write_inline_data(struct inode *inode,
				  unsigned len,
				  struct page *page)
{
	int ret;
	void *kaddr;
	struct ldiskfs_iloc iloc;

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret) {
		ldiskfs_std_error(inode->i_sb, ret);
		return NULL;
	}

	down_write(&LDISKFS_I(inode)->xattr_sem);
	kaddr = kmap_atomic(page);
	ldiskfs_write_inline_data(inode, &iloc, kaddr, 0, len);
	kunmap_atomic(kaddr);
	up_write(&LDISKFS_I(inode)->xattr_sem);

	return iloc.bh;
}

/*
 * Try to make the page cache and handle ready for the inline data case.
 * We can call this function in 2 cases:
 * 1. The inode is created and the first write exceeds inline size. We can
 *    clear the inode state safely.
 * 2. The inode has inline data, then we need to read the data, make it
 *    update and dirty so that ldiskfs_da_writepages can handle it. We don't
 *    need to start the journal since the file's metatdata isn't changed now.
 */
static int ldiskfs_da_convert_inline_data_to_extent(struct address_space *mapping,
						 struct inode *inode,
						 unsigned flags,
						 void **fsdata)
{
	int ret = 0, inline_size;
	struct page *page;

	page = grab_cache_page_write_begin(mapping, 0, flags);
	if (!page)
		return -ENOMEM;

	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		ldiskfs_clear_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
		goto out;
	}

	inline_size = ldiskfs_get_inline_size(inode);

	if (!PageUptodate(page)) {
		ret = ldiskfs_read_inline_page(inode, page);
		if (ret < 0)
			goto out;
	}

	ret = __block_write_begin(page, 0, inline_size,
				  ldiskfs_da_get_block_prep);
	if (ret) {
		up_read(&LDISKFS_I(inode)->xattr_sem);
		unlock_page(page);
		page_cache_release(page);
		ldiskfs_truncate_failed_write(inode);
		return ret;
	}

	SetPageDirty(page);
	SetPageUptodate(page);
	ldiskfs_clear_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
	*fsdata = (void *)CONVERT_INLINE_DATA;

out:
	up_read(&LDISKFS_I(inode)->xattr_sem);
	if (page) {
		unlock_page(page);
		page_cache_release(page);
	}
	return ret;
}

/*
 * Prepare the write for the inline data.
 * If the the data can be written into the inode, we just read
 * the page and make it uptodate, and start the journal.
 * Otherwise read the page, makes it dirty so that it can be
 * handle in writepages(the i_disksize update is left to the
 * normal ldiskfs_da_write_end).
 */
int ldiskfs_da_write_inline_data_begin(struct address_space *mapping,
				    struct inode *inode,
				    loff_t pos, unsigned len,
				    unsigned flags,
				    struct page **pagep,
				    void **fsdata)
{
	int ret, inline_size;
	handle_t *handle;
	struct page *page;
	struct ldiskfs_iloc iloc;
	int retries = 0;

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		return ret;

retry_journal:
	handle = ldiskfs_journal_start(inode, LDISKFS_HT_INODE, 1);
	if (IS_ERR(handle)) {
		ret = PTR_ERR(handle);
		goto out;
	}

	inline_size = ldiskfs_get_max_inline_size(inode);

	ret = -ENOSPC;
	if (inline_size >= pos + len) {
		ret = ldiskfs_prepare_inline_data(handle, inode, pos + len);
		if (ret && ret != -ENOSPC)
			goto out_journal;
	}

	/*
	 * We cannot recurse into the filesystem as the transaction
	 * is already started.
	 */
	flags |= AOP_FLAG_NOFS;

	if (ret == -ENOSPC) {
		ldiskfs_journal_stop(handle);
		ret = ldiskfs_da_convert_inline_data_to_extent(mapping,
							    inode,
							    flags,
							    fsdata);
		if (ret == -ENOSPC &&
		    ldiskfs_should_retry_alloc(inode->i_sb, &retries))
			goto retry_journal;
		goto out;
	}


	page = grab_cache_page_write_begin(mapping, 0, flags);
	if (!page) {
		ret = -ENOMEM;
		goto out_journal;
	}

	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		ret = 0;
		goto out_release_page;
	}

	if (!PageUptodate(page)) {
		ret = ldiskfs_read_inline_page(inode, page);
		if (ret < 0)
			goto out_release_page;
	}

	up_read(&LDISKFS_I(inode)->xattr_sem);
	*pagep = page;
	brelse(iloc.bh);
	return 1;
out_release_page:
	up_read(&LDISKFS_I(inode)->xattr_sem);
	unlock_page(page);
	page_cache_release(page);
out_journal:
	ldiskfs_journal_stop(handle);
out:
	brelse(iloc.bh);
	return ret;
}

int ldiskfs_da_write_inline_data_end(struct inode *inode, loff_t pos,
				  unsigned len, unsigned copied,
				  struct page *page)
{
	int i_size_changed = 0;

	copied = ldiskfs_write_inline_data_end(inode, pos, len, copied, page);

	/*
	 * No need to use i_size_read() here, the i_size
	 * cannot change under us because we hold i_mutex.
	 *
	 * But it's important to update i_size while still holding page lock:
	 * page writeout could otherwise come in and zero beyond i_size.
	 */
	if (pos+copied > inode->i_size) {
		i_size_write(inode, pos+copied);
		i_size_changed = 1;
	}
	unlock_page(page);
	page_cache_release(page);

	/*
	 * Don't mark the inode dirty under page lock. First, it unnecessarily
	 * makes the holding time of page lock longer. Second, it forces lock
	 * ordering of page lock and transaction start for journaling
	 * filesystems.
	 */
	if (i_size_changed)
		mark_inode_dirty(inode);

	return copied;
}

#ifdef INLINE_DIR_DEBUG
void ldiskfs_show_inline_dir(struct inode *dir, struct buffer_head *bh,
			  void *inline_start, int inline_size)
{
	int offset;
	unsigned short de_len;
	struct ldiskfs_dir_entry_2 *de = inline_start;
	void *dlimit = inline_start + inline_size;

	trace_printk("inode %lu\n", dir->i_ino);
	offset = 0;
	while ((void *)de < dlimit) {
		de_len = ldiskfs_rec_len_from_disk(de->rec_len, inline_size);
		trace_printk("de: off %u rlen %u name %.*s nlen %u ino %u\n",
			     offset, de_len, de->name_len, de->name,
			     de->name_len, le32_to_cpu(de->inode));
		if (ldiskfs_check_dir_entry(dir, NULL, de, bh,
					 inline_start, inline_size, offset))
			BUG();

		offset += de_len;
		de = (struct ldiskfs_dir_entry_2 *) ((char *) de + de_len);
	}
}
#else
#define ldiskfs_show_inline_dir(dir, bh, inline_start, inline_size)
#endif

/*
 * Add a new entry into a inline dir.
 * It will return -ENOSPC if no space is available, and -EIO
 * and -EEXIST if directory entry already exists.
 */
static int ldiskfs_add_dirent_to_inline(handle_t *handle,
				     struct dentry *dentry,
				     struct inode *inode,
				     struct ldiskfs_iloc *iloc,
				     void *inline_start, int inline_size)
{
	struct inode	*dir = dentry->d_parent->d_inode;
	const char	*name = dentry->d_name.name;
	int		namelen = dentry->d_name.len;
	int		err;
	struct ldiskfs_dir_entry_2 *de;

	err = ldiskfs_find_dest_de(dir, inode, iloc->bh,
				inline_start, inline_size,
				name, namelen, &de, NULL);
	if (err)
		return err;

	BUFFER_TRACE(iloc->bh, "get_write_access");
	err = ldiskfs_journal_get_write_access(handle, iloc->bh);
	if (err)
		return err;
	ldiskfs_insert_dentry(inode, de, inline_size, name, namelen, NULL);

	ldiskfs_show_inline_dir(dir, iloc->bh, inline_start, inline_size);

	/*
	 * XXX shouldn't update any times until successful
	 * completion of syscall, but too many callers depend
	 * on this.
	 *
	 * XXX similarly, too many callers depend on
	 * ldiskfs_new_inode() setting the times, but error
	 * recovery deletes the inode, so the worst that can
	 * happen is that the times are slightly out of date
	 * and/or different from the directory change time.
	 */
	dir->i_mtime = dir->i_ctime = ldiskfs_current_time(dir);
	ldiskfs_update_dx_flag(dir);
	dir->i_version++;
	ldiskfs_mark_inode_dirty(handle, dir);
	return 1;
}

static void *ldiskfs_get_inline_xattr_pos(struct inode *inode,
				       struct ldiskfs_iloc *iloc)
{
	struct ldiskfs_xattr_entry *entry;
	struct ldiskfs_xattr_ibody_header *header;

	BUG_ON(!LDISKFS_I(inode)->i_inline_off);

	header = IHDR(inode, ldiskfs_raw_inode(iloc));
	entry = (struct ldiskfs_xattr_entry *)((void *)ldiskfs_raw_inode(iloc) +
					    LDISKFS_I(inode)->i_inline_off);

	return (void *)IFIRST(header) + le16_to_cpu(entry->e_value_offs);
}

/* Set the final de to cover the whole block. */
static void ldiskfs_update_final_de(void *de_buf, int old_size, int new_size)
{
	struct ldiskfs_dir_entry_2 *de, *prev_de;
	void *limit;
	int de_len;

	de = (struct ldiskfs_dir_entry_2 *)de_buf;
	if (old_size) {
		limit = de_buf + old_size;
		do {
			prev_de = de;
			de_len = ldiskfs_rec_len_from_disk(de->rec_len, old_size);
			de_buf += de_len;
			de = (struct ldiskfs_dir_entry_2 *)de_buf;
		} while (de_buf < limit);

		prev_de->rec_len = ldiskfs_rec_len_to_disk(de_len + new_size -
							old_size, new_size);
	} else {
		/* this is just created, so create an empty entry. */
		de->inode = 0;
		de->rec_len = ldiskfs_rec_len_to_disk(new_size, new_size);
	}
}

static int ldiskfs_update_inline_dir(handle_t *handle, struct inode *dir,
				  struct ldiskfs_iloc *iloc)
{
	int ret;
	int old_size = LDISKFS_I(dir)->i_inline_size - LDISKFS_MIN_INLINE_DATA_SIZE;
	int new_size = get_max_inline_xattr_value_size(dir, iloc);

	if (new_size - old_size <= __LDISKFS_DIR_REC_LEN(1))
		return -ENOSPC;

	ret = ldiskfs_update_inline_data(handle, dir,
				      new_size + LDISKFS_MIN_INLINE_DATA_SIZE);
	if (ret)
		return ret;

	ldiskfs_update_final_de(ldiskfs_get_inline_xattr_pos(dir, iloc), old_size,
			     LDISKFS_I(dir)->i_inline_size -
						LDISKFS_MIN_INLINE_DATA_SIZE);
	dir->i_size = LDISKFS_I(dir)->i_disksize = LDISKFS_I(dir)->i_inline_size;
	return 0;
}

static void ldiskfs_restore_inline_data(handle_t *handle, struct inode *inode,
				     struct ldiskfs_iloc *iloc,
				     void *buf, int inline_size)
{
	ldiskfs_create_inline_data(handle, inode, inline_size);
	ldiskfs_write_inline_data(inode, iloc, buf, 0, inline_size);
	ldiskfs_set_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
}

static int ldiskfs_finish_convert_inline_dir(handle_t *handle,
					  struct inode *inode,
					  struct buffer_head *dir_block,
					  void *buf,
					  int inline_size)
{
	int err, csum_size = 0, header_size = 0;
	struct ldiskfs_dir_entry_2 *de;
	struct ldiskfs_dir_entry_tail *t;
	void *target = dir_block->b_data;

	/*
	 * First create "." and ".." and then copy the dir information
	 * back to the block.
	 */
	de = (struct ldiskfs_dir_entry_2 *)target;
	de = ldiskfs_init_dot_dotdot(inode, de,
		inode->i_sb->s_blocksize, csum_size,
		le32_to_cpu(((struct ldiskfs_dir_entry_2 *)buf)->inode), 1);
	header_size = (void *)de - target;

	memcpy((void *)de, buf + LDISKFS_INLINE_DOTDOT_SIZE,
		inline_size - LDISKFS_INLINE_DOTDOT_SIZE);

	if (ldiskfs_has_metadata_csum(inode->i_sb))
		csum_size = sizeof(struct ldiskfs_dir_entry_tail);

	inode->i_size = inode->i_sb->s_blocksize;
	i_size_write(inode, inode->i_sb->s_blocksize);
	LDISKFS_I(inode)->i_disksize = inode->i_sb->s_blocksize;
	ldiskfs_update_final_de(dir_block->b_data,
			inline_size - LDISKFS_INLINE_DOTDOT_SIZE + header_size,
			inode->i_sb->s_blocksize - csum_size);

	if (csum_size) {
		t = LDISKFS_DIRENT_TAIL(dir_block->b_data,
				     inode->i_sb->s_blocksize);
		initialize_dirent_tail(t, inode->i_sb->s_blocksize);
	}
	set_buffer_uptodate(dir_block);
	err = ldiskfs_handle_dirty_dirent_node(handle, inode, dir_block);
	if (err)
		return err;
	set_buffer_verified(dir_block);
	return ldiskfs_mark_inode_dirty(handle, inode);
}

static int ldiskfs_convert_inline_data_nolock(handle_t *handle,
					   struct inode *inode,
					   struct ldiskfs_iloc *iloc)
{
	int error;
	void *buf = NULL;
	struct buffer_head *data_bh = NULL;
	struct ldiskfs_map_blocks map;
	int inline_size;

	inline_size = ldiskfs_get_inline_size(inode);
	buf = kmalloc(inline_size, GFP_NOFS);
	if (!buf) {
		error = -ENOMEM;
		goto out;
	}

	error = ldiskfs_read_inline_data(inode, buf, inline_size, iloc);
	if (error < 0)
		goto out;

	/*
	 * Make sure the inline directory entries pass checks before we try to
	 * convert them, so that we avoid touching stuff that needs fsck.
	 */
	if (S_ISDIR(inode->i_mode)) {
		error = ldiskfs_check_all_de(inode, iloc->bh,
					buf + LDISKFS_INLINE_DOTDOT_SIZE,
					inline_size - LDISKFS_INLINE_DOTDOT_SIZE);
		if (error)
			goto out;
	}

	error = ldiskfs_destroy_inline_data_nolock(handle, inode);
	if (error)
		goto out;

	map.m_lblk = 0;
	map.m_len = 1;
	map.m_flags = 0;
	error = ldiskfs_map_blocks(handle, inode, &map, LDISKFS_GET_BLOCKS_CREATE);
	if (error < 0)
		goto out_restore;
	if (!(map.m_flags & LDISKFS_MAP_MAPPED)) {
		error = -EIO;
		goto out_restore;
	}

	data_bh = sb_getblk(inode->i_sb, map.m_pblk);
	if (!data_bh) {
		error = -ENOMEM;
		goto out_restore;
	}

	lock_buffer(data_bh);
	error = ldiskfs_journal_get_create_access(handle, data_bh);
	if (error) {
		unlock_buffer(data_bh);
		error = -EIO;
		goto out_restore;
	}
	memset(data_bh->b_data, 0, inode->i_sb->s_blocksize);

	if (!S_ISDIR(inode->i_mode)) {
		memcpy(data_bh->b_data, buf, inline_size);
		set_buffer_uptodate(data_bh);
		error = ldiskfs_handle_dirty_metadata(handle,
						   inode, data_bh);
	} else {
		error = ldiskfs_finish_convert_inline_dir(handle, inode, data_bh,
						       buf, inline_size);
	}

	unlock_buffer(data_bh);
out_restore:
	if (error)
		ldiskfs_restore_inline_data(handle, inode, iloc, buf, inline_size);

out:
	brelse(data_bh);
	kfree(buf);
	return error;
}

/*
 * Try to add the new entry to the inline data.
 * If succeeds, return 0. If not, extended the inline dir and copied data to
 * the new created block.
 */
int ldiskfs_try_add_inline_entry(handle_t *handle, struct dentry *dentry,
			      struct inode *inode)
{
	int ret, inline_size;
	void *inline_start;
	struct ldiskfs_iloc iloc;
	struct inode *dir = dentry->d_parent->d_inode;

	ret = ldiskfs_get_inode_loc(dir, &iloc);
	if (ret)
		return ret;

	down_write(&LDISKFS_I(dir)->xattr_sem);
	if (!ldiskfs_has_inline_data(dir))
		goto out;

	inline_start = (void *)ldiskfs_raw_inode(&iloc)->i_block +
						 LDISKFS_INLINE_DOTDOT_SIZE;
	inline_size = LDISKFS_MIN_INLINE_DATA_SIZE - LDISKFS_INLINE_DOTDOT_SIZE;

	ret = ldiskfs_add_dirent_to_inline(handle, dentry, inode, &iloc,
					inline_start, inline_size);
	if (ret != -ENOSPC)
		goto out;

	/* check whether it can be inserted to inline xattr space. */
	inline_size = LDISKFS_I(dir)->i_inline_size -
			LDISKFS_MIN_INLINE_DATA_SIZE;
	if (!inline_size) {
		/* Try to use the xattr space.*/
		ret = ldiskfs_update_inline_dir(handle, dir, &iloc);
		if (ret && ret != -ENOSPC)
			goto out;

		inline_size = LDISKFS_I(dir)->i_inline_size -
				LDISKFS_MIN_INLINE_DATA_SIZE;
	}

	if (inline_size) {
		inline_start = ldiskfs_get_inline_xattr_pos(dir, &iloc);

		ret = ldiskfs_add_dirent_to_inline(handle, dentry, inode, &iloc,
						inline_start, inline_size);

		if (ret != -ENOSPC)
			goto out;
	}

	/*
	 * The inline space is filled up, so create a new block for it.
	 * As the extent tree will be created, we have to save the inline
	 * dir first.
	 */
	ret = ldiskfs_convert_inline_data_nolock(handle, dir, &iloc);

out:
	ldiskfs_mark_inode_dirty(handle, dir);
	up_write(&LDISKFS_I(dir)->xattr_sem);
	brelse(iloc.bh);
	return ret;
}

/*
 * This function fills a red-black tree with information from an
 * inlined dir.  It returns the number directory entries loaded
 * into the tree.  If there is an error it is returned in err.
 */
int htree_inlinedir_to_tree(struct file *dir_file,
			    struct inode *dir, ldiskfs_lblk_t block,
			    struct ldiskfs_dx_hash_info *hinfo,
			    __u32 start_hash, __u32 start_minor_hash,
			    int *has_inline_data)
{
	int err = 0, count = 0;
	unsigned int parent_ino;
	int pos;
	struct ldiskfs_dir_entry_2 *de;
	struct inode *inode = file_inode(dir_file);
	int ret, inline_size = 0;
	struct ldiskfs_iloc iloc;
	void *dir_buf = NULL;
	struct ldiskfs_dir_entry_2 fake;

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		return ret;

	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		up_read(&LDISKFS_I(inode)->xattr_sem);
		*has_inline_data = 0;
		goto out;
	}

	inline_size = ldiskfs_get_inline_size(inode);
	dir_buf = kmalloc(inline_size, GFP_NOFS);
	if (!dir_buf) {
		ret = -ENOMEM;
		up_read(&LDISKFS_I(inode)->xattr_sem);
		goto out;
	}

	ret = ldiskfs_read_inline_data(inode, dir_buf, inline_size, &iloc);
	up_read(&LDISKFS_I(inode)->xattr_sem);
	if (ret < 0)
		goto out;

	pos = 0;
	parent_ino = le32_to_cpu(((struct ldiskfs_dir_entry_2 *)dir_buf)->inode);
	while (pos < inline_size) {
		/*
		 * As inlined dir doesn't store any information about '.' and
		 * only the inode number of '..' is stored, we have to handle
		 * them differently.
		 */
		if (pos == 0) {
			fake.inode = cpu_to_le32(inode->i_ino);
			fake.name_len = 1;
			strcpy(fake.name, ".");
			fake.rec_len = ldiskfs_rec_len_to_disk(
						LDISKFS_DIR_REC_LEN(&fake),
						inline_size);
			ldiskfs_set_de_type(inode->i_sb, &fake, S_IFDIR);
			de = &fake;
			pos = LDISKFS_INLINE_DOTDOT_OFFSET;
		} else if (pos == LDISKFS_INLINE_DOTDOT_OFFSET) {
			fake.inode = cpu_to_le32(parent_ino);
			fake.name_len = 2;
			strcpy(fake.name, "..");
			fake.rec_len = ldiskfs_rec_len_to_disk(
						LDISKFS_DIR_REC_LEN(&fake),
						inline_size);
			ldiskfs_set_de_type(inode->i_sb, &fake, S_IFDIR);
			de = &fake;
			pos = LDISKFS_INLINE_DOTDOT_SIZE;
		} else {
			de = (struct ldiskfs_dir_entry_2 *)(dir_buf + pos);
			pos += ldiskfs_rec_len_from_disk(de->rec_len, inline_size);
			if (ldiskfs_check_dir_entry(inode, dir_file, de,
					 iloc.bh, dir_buf,
					 inline_size, pos)) {
				ret = count;
				goto out;
			}
		}

		ldiskfsfs_dirhash(de->name, de->name_len, hinfo);
		if ((hinfo->hash < start_hash) ||
		    ((hinfo->hash == start_hash) &&
		     (hinfo->minor_hash < start_minor_hash)))
			continue;
		if (de->inode == 0)
			continue;
		err = ldiskfs_htree_store_dirent(dir_file,
				   hinfo->hash, hinfo->minor_hash, de);
		if (err) {
			count = err;
			goto out;
		}
		count++;
	}
	ret = count;
out:
	kfree(dir_buf);
	brelse(iloc.bh);
	return ret;
}

/*
 * So this function is called when the volume is mkfsed with
 * dir_index disabled. In order to keep f_pos persistent
 * after we convert from an inlined dir to a blocked based,
 * we just pretend that we are a normal dir and return the
 * offset as if '.' and '..' really take place.
 *
 */
int ldiskfs_read_inline_dir(struct file *filp,
			 void *dirent, filldir_t filldir,
			 int *has_inline_data)
{
	int error = 0;
	unsigned int offset, parent_ino;
	int i, stored;
	struct ldiskfs_dir_entry_2 *de;
	struct super_block *sb;
	struct inode *inode = file_inode(filp);
	int ret, inline_size = 0;
	struct ldiskfs_iloc iloc;
	void *dir_buf = NULL;
	int dotdot_offset, dotdot_size, extra_offset, extra_size;

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		return ret;

	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		up_read(&LDISKFS_I(inode)->xattr_sem);
		*has_inline_data = 0;
		goto out;
	}

	inline_size = ldiskfs_get_inline_size(inode);
	dir_buf = kmalloc(inline_size, GFP_NOFS);
	if (!dir_buf) {
		ret = -ENOMEM;
		up_read(&LDISKFS_I(inode)->xattr_sem);
		goto out;
	}

	ret = ldiskfs_read_inline_data(inode, dir_buf, inline_size, &iloc);
	up_read(&LDISKFS_I(inode)->xattr_sem);
	if (ret < 0)
		goto out;

	ret = 0;
	sb = inode->i_sb;
	stored = 0;
	parent_ino = le32_to_cpu(((struct ldiskfs_dir_entry_2 *)dir_buf)->inode);
	offset = filp->f_pos;

	/*
	 * dotdot_offset and dotdot_size is the real offset and
	 * size for ".." and "." if the dir is block based while
	 * the real size for them are only LDISKFS_INLINE_DOTDOT_SIZE.
	 * So we will use extra_offset and extra_size to indicate them
	 * during the inline dir iteration.
	 */
	dotdot_offset = __LDISKFS_DIR_REC_LEN(1);
	dotdot_size = dotdot_offset + __LDISKFS_DIR_REC_LEN(2);
	extra_offset = dotdot_size - LDISKFS_INLINE_DOTDOT_SIZE;
	extra_size = extra_offset + inline_size;

	while (!error && !stored && filp->f_pos < extra_size) {
revalidate:
		/*
		 * If the version has changed since the last call to
		 * readdir(2), then we might be pointing to an invalid
		 * dirent right now.  Scan from the start of the inline
		 * dir to make sure.
		 */
		if (filp->f_version != inode->i_version) {
			for (i = 0; i < extra_size && i < offset;) {
				/*
				 * "." is with offset 0 and
				 * ".." is dotdot_offset.
				 */
				if (!i) {
					i = dotdot_offset;
					continue;
				} else if (i == dotdot_offset) {
					i = dotdot_size;
					continue;
				}
				/* for other entry, the real offset in
				 * the buf has to be tuned accordingly.
				 */
				de = (struct ldiskfs_dir_entry_2 *)
					(dir_buf + i - extra_offset);
				/* It's too expensive to do a full
				 * dirent test each time round this
				 * loop, but we do have to test at
				 * least that it is non-zero.  A
				 * failure will be detected in the
				 * dirent test below. */
				if (ldiskfs_rec_len_from_disk(de->rec_len,
					extra_size) < __LDISKFS_DIR_REC_LEN(1))
					break;
				i += ldiskfs_rec_len_from_disk(de->rec_len,
							    extra_size);
			}
			offset = i;
			filp->f_pos = offset;
			filp->f_version = inode->i_version;
		}

		while (!error && filp->f_pos < extra_size) {
			if (filp->f_pos == 0) {
				error = filldir(dirent, ".", 1, 0, inode->i_ino,
						DT_DIR);
				if (error)
					break;
				stored++;
				filp->f_pos = dotdot_offset;
				continue;
			}

			if (filp->f_pos == dotdot_offset) {
				error = filldir(dirent, "..", 2,
						dotdot_offset,
						parent_ino, DT_DIR);
				if (error)
					break;
				stored++;

				filp->f_pos = dotdot_size;
				continue;
			}

			de = (struct ldiskfs_dir_entry_2 *)
				(dir_buf + filp->f_pos - extra_offset);
			if (ldiskfs_check_dir_entry(inode, filp, de,
						 iloc.bh, dir_buf,
						 extra_size, filp->f_pos)) {
				ret = stored;
				goto out;
			}
			if (le32_to_cpu(de->inode)) {
				/* We might block in the next section
				 * if the data destination is
				 * currently swapped out.  So, use a
				 * version stamp to detect whether or
				 * not the directory has been modified
				 * during the copy operation.
				 */
				u64 version = filp->f_version;

				error = filldir(dirent, de->name,
						de->name_len,
						filp->f_pos,
						le32_to_cpu(de->inode),
						get_dtype(sb, de->file_type));
				if (error)
					break;
				if (version != filp->f_version)
					goto revalidate;
				stored++;
			}
			filp->f_pos += ldiskfs_rec_len_from_disk(de->rec_len,
							      extra_size);
		}
	}
out:
	kfree(dir_buf);
	brelse(iloc.bh);
	return ret;
}

struct buffer_head *ldiskfs_get_first_inline_block(struct inode *inode,
					struct ldiskfs_dir_entry_2 **parent_de,
					int *retval)
{
	struct ldiskfs_iloc iloc;

	*retval = ldiskfs_get_inode_loc(inode, &iloc);
	if (*retval)
		return NULL;

	*parent_de = (struct ldiskfs_dir_entry_2 *)ldiskfs_raw_inode(&iloc)->i_block;

	return iloc.bh;
}

/*
 * Try to create the inline data for the new dir.
 * If it succeeds, return 0, otherwise return the error.
 * In case of ENOSPC, the caller should create the normal disk layout dir.
 */
int ldiskfs_try_create_inline_dir(handle_t *handle, struct inode *parent,
			       struct inode *inode)
{
	int ret, inline_size = LDISKFS_MIN_INLINE_DATA_SIZE;
	struct ldiskfs_iloc iloc;
	struct ldiskfs_dir_entry_2 *de;

	ret = ldiskfs_get_inode_loc(inode, &iloc);
	if (ret)
		return ret;

	ret = ldiskfs_prepare_inline_data(handle, inode, inline_size);
	if (ret)
		goto out;

	/*
	 * For inline dir, we only save the inode information for the ".."
	 * and create a fake dentry to cover the left space.
	 */
	de = (struct ldiskfs_dir_entry_2 *)ldiskfs_raw_inode(&iloc)->i_block;
	de->inode = cpu_to_le32(parent->i_ino);
	de = (struct ldiskfs_dir_entry_2 *)((void *)de + LDISKFS_INLINE_DOTDOT_SIZE);
	de->inode = 0;
	de->rec_len = ldiskfs_rec_len_to_disk(
				inline_size - LDISKFS_INLINE_DOTDOT_SIZE,
				inline_size);
	set_nlink(inode, 2);
	inode->i_size = LDISKFS_I(inode)->i_disksize = inline_size;
out:
	brelse(iloc.bh);
	return ret;
}

struct buffer_head *ldiskfs_find_inline_entry(struct inode *dir,
					const struct qstr *d_name,
					struct ldiskfs_dir_entry_2 **res_dir,
					int *has_inline_data)
{
	int ret;
	struct ldiskfs_iloc iloc;
	void *inline_start;
	int inline_size;

	if (ldiskfs_get_inode_loc(dir, &iloc))
		return NULL;

	down_read(&LDISKFS_I(dir)->xattr_sem);
	if (!ldiskfs_has_inline_data(dir)) {
		*has_inline_data = 0;
		goto out;
	}

	inline_start = (void *)ldiskfs_raw_inode(&iloc)->i_block +
						LDISKFS_INLINE_DOTDOT_SIZE;
	inline_size = LDISKFS_MIN_INLINE_DATA_SIZE - LDISKFS_INLINE_DOTDOT_SIZE;
	ret = search_dir(iloc.bh, inline_start, inline_size,
			 dir, d_name, 0, res_dir);
	if (ret == 1)
		goto out_find;
	if (ret < 0)
		goto out;

	if (ldiskfs_get_inline_size(dir) == LDISKFS_MIN_INLINE_DATA_SIZE)
		goto out;

	inline_start = ldiskfs_get_inline_xattr_pos(dir, &iloc);
	inline_size = ldiskfs_get_inline_size(dir) - LDISKFS_MIN_INLINE_DATA_SIZE;

	ret = search_dir(iloc.bh, inline_start, inline_size,
			 dir, d_name, 0, res_dir);
	if (ret == 1)
		goto out_find;

out:
	brelse(iloc.bh);
	iloc.bh = NULL;
out_find:
	up_read(&LDISKFS_I(dir)->xattr_sem);
	return iloc.bh;
}

int ldiskfs_delete_inline_entry(handle_t *handle,
			     struct inode *dir,
			     struct ldiskfs_dir_entry_2 *de_del,
			     struct buffer_head *bh,
			     int *has_inline_data)
{
	int err, inline_size;
	struct ldiskfs_iloc iloc;
	void *inline_start;

	err = ldiskfs_get_inode_loc(dir, &iloc);
	if (err)
		return err;

	down_write(&LDISKFS_I(dir)->xattr_sem);
	if (!ldiskfs_has_inline_data(dir)) {
		*has_inline_data = 0;
		goto out;
	}

	if ((void *)de_del - ((void *)ldiskfs_raw_inode(&iloc)->i_block) <
		LDISKFS_MIN_INLINE_DATA_SIZE) {
		inline_start = (void *)ldiskfs_raw_inode(&iloc)->i_block +
					LDISKFS_INLINE_DOTDOT_SIZE;
		inline_size = LDISKFS_MIN_INLINE_DATA_SIZE -
				LDISKFS_INLINE_DOTDOT_SIZE;
	} else {
		inline_start = ldiskfs_get_inline_xattr_pos(dir, &iloc);
		inline_size = ldiskfs_get_inline_size(dir) -
				LDISKFS_MIN_INLINE_DATA_SIZE;
	}

	BUFFER_TRACE(bh, "get_write_access");
	err = ldiskfs_journal_get_write_access(handle, bh);
	if (err)
		goto out;

	err = ldiskfs_generic_delete_entry(handle, dir, de_del, bh,
					inline_start, inline_size, 0);
	if (err)
		goto out;

	BUFFER_TRACE(bh, "call ldiskfs_handle_dirty_metadata");
	err = ldiskfs_mark_inode_dirty(handle, dir);
	if (unlikely(err))
		goto out;

	ldiskfs_show_inline_dir(dir, iloc.bh, inline_start, inline_size);
out:
	up_write(&LDISKFS_I(dir)->xattr_sem);
	brelse(iloc.bh);
	if (err != -ENOENT)
		ldiskfs_std_error(dir->i_sb, err);
	return err;
}

/*
 * Get the inline dentry at offset.
 */
static inline struct ldiskfs_dir_entry_2 *
ldiskfs_get_inline_entry(struct inode *inode,
		      struct ldiskfs_iloc *iloc,
		      unsigned int offset,
		      void **inline_start,
		      int *inline_size)
{
	void *inline_pos;

	BUG_ON(offset > ldiskfs_get_inline_size(inode));

	if (offset < LDISKFS_MIN_INLINE_DATA_SIZE) {
		inline_pos = (void *)ldiskfs_raw_inode(iloc)->i_block;
		*inline_size = LDISKFS_MIN_INLINE_DATA_SIZE;
	} else {
		inline_pos = ldiskfs_get_inline_xattr_pos(inode, iloc);
		offset -= LDISKFS_MIN_INLINE_DATA_SIZE;
		*inline_size = ldiskfs_get_inline_size(inode) -
				LDISKFS_MIN_INLINE_DATA_SIZE;
	}

	if (inline_start)
		*inline_start = inline_pos;
	return (struct ldiskfs_dir_entry_2 *)(inline_pos + offset);
}

int empty_inline_dir(struct inode *dir, int *has_inline_data)
{
	int err, inline_size;
	struct ldiskfs_iloc iloc;
	size_t inline_len;
	void *inline_pos;
	unsigned int offset;
	struct ldiskfs_dir_entry_2 *de;
	int ret = 1;

	err = ldiskfs_get_inode_loc(dir, &iloc);
	if (err) {
		LDISKFS_ERROR_INODE(dir, "error %d getting inode %lu block",
				 err, dir->i_ino);
		return 1;
	}

	down_read(&LDISKFS_I(dir)->xattr_sem);
	if (!ldiskfs_has_inline_data(dir)) {
		*has_inline_data = 0;
		goto out;
	}

	de = (struct ldiskfs_dir_entry_2 *)ldiskfs_raw_inode(&iloc)->i_block;
	if (!le32_to_cpu(de->inode)) {
		ldiskfs_warning(dir->i_sb,
			     "bad inline directory (dir #%lu) - no `..'",
			     dir->i_ino);
		ret = 1;
		goto out;
	}

	inline_len = ldiskfs_get_inline_size(dir);
	offset = LDISKFS_INLINE_DOTDOT_SIZE;
	while (offset < inline_len) {
		de = ldiskfs_get_inline_entry(dir, &iloc, offset,
					   &inline_pos, &inline_size);
		if (ldiskfs_check_dir_entry(dir, NULL, de,
					 iloc.bh, inline_pos,
					 inline_size, offset)) {
			ldiskfs_warning(dir->i_sb,
				     "bad inline directory (dir #%lu) - "
				     "inode %u, rec_len %u, name_len %d"
				     "inline size %d\n",
				     dir->i_ino, le32_to_cpu(de->inode),
				     le16_to_cpu(de->rec_len), de->name_len,
				     inline_size);
			ret = 1;
			goto out;
		}
		if (le32_to_cpu(de->inode)) {
			ret = 0;
			goto out;
		}
		offset += ldiskfs_rec_len_from_disk(de->rec_len, inline_size);
	}

out:
	up_read(&LDISKFS_I(dir)->xattr_sem);
	brelse(iloc.bh);
	return ret;
}

int ldiskfs_destroy_inline_data(handle_t *handle, struct inode *inode)
{
	int ret;

	down_write(&LDISKFS_I(inode)->xattr_sem);
	ret = ldiskfs_destroy_inline_data_nolock(handle, inode);
	up_write(&LDISKFS_I(inode)->xattr_sem);

	return ret;
}

int ldiskfs_inline_data_fiemap(struct inode *inode,
			    struct fiemap_extent_info *fieinfo,
			    int *has_inline, __u64 start, __u64 len)
{
	__u64 physical = 0;
	__u64 inline_len;
	__u32 flags = FIEMAP_EXTENT_DATA_INLINE | FIEMAP_EXTENT_NOT_ALIGNED |
		FIEMAP_EXTENT_LAST;
	int error = 0;
	struct ldiskfs_iloc iloc;

	down_read(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		*has_inline = 0;
		goto out;
	}
	inline_len = min_t(size_t, ldiskfs_get_inline_size(inode),
			   i_size_read(inode));
	if (start >= inline_len)
		goto out;
	if (start + len < inline_len)
		inline_len = start + len;
	inline_len -= start;

	error = ldiskfs_get_inode_loc(inode, &iloc);
	if (error)
		goto out;

	physical = (__u64)iloc.bh->b_blocknr << inode->i_sb->s_blocksize_bits;
	physical += (char *)ldiskfs_raw_inode(&iloc) - iloc.bh->b_data;
	physical += offsetof(struct ldiskfs_inode, i_block);

	if (physical)
		error = fiemap_fill_next_extent(fieinfo, start, physical,
						inline_len, flags);
	brelse(iloc.bh);
out:
	up_read(&LDISKFS_I(inode)->xattr_sem);
	return (error < 0 ? error : 0);
}

void ldiskfs_inline_data_truncate(struct inode *inode, int *has_inline)
{
	handle_t *handle;
	int inline_size, value_len, needed_blocks;
	size_t i_size;
	void *value = NULL;
	struct ldiskfs_xattr_ibody_find is = {
		.s = { .not_found = -ENODATA, },
	};
	struct ldiskfs_xattr_info i = {
		.name_index = LDISKFS_XATTR_INDEX_SYSTEM,
		.name = LDISKFS_XATTR_SYSTEM_DATA,
	};


	needed_blocks = ldiskfs_writepage_trans_blocks(inode);
	handle = ldiskfs_journal_start(inode, LDISKFS_HT_INODE, needed_blocks);
	if (IS_ERR(handle))
		return;

	down_write(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		*has_inline = 0;
		ldiskfs_journal_stop(handle);
		return;
	}

	if (ldiskfs_orphan_add(handle, inode))
		goto out;

	if (ldiskfs_get_inode_loc(inode, &is.iloc))
		goto out;

	down_write(&LDISKFS_I(inode)->i_data_sem);
	i_size = inode->i_size;
	inline_size = ldiskfs_get_inline_size(inode);
	LDISKFS_I(inode)->i_disksize = i_size;

	if (i_size < inline_size) {
		/* Clear the content in the xattr space. */
		if (inline_size > LDISKFS_MIN_INLINE_DATA_SIZE) {
			if (ldiskfs_xattr_ibody_find(inode, &i, &is))
				goto out_error;

			BUG_ON(is.s.not_found);

			value_len = le32_to_cpu(is.s.here->e_value_size);
			value = kmalloc(value_len, GFP_NOFS);
			if (!value)
				goto out_error;

			if (ldiskfs_xattr_ibody_get(inode, i.name_index, i.name,
						value, value_len))
				goto out_error;

			i.value = value;
			i.value_len = i_size > LDISKFS_MIN_INLINE_DATA_SIZE ?
					i_size - LDISKFS_MIN_INLINE_DATA_SIZE : 0;
			if (ldiskfs_xattr_ibody_inline_set(handle, inode, &i, &is))
				goto out_error;
		}

		/* Clear the content within i_blocks. */
		if (i_size < LDISKFS_MIN_INLINE_DATA_SIZE) {
			void *p = (void *) ldiskfs_raw_inode(&is.iloc)->i_block;
			memset(p + i_size, 0,
			       LDISKFS_MIN_INLINE_DATA_SIZE - i_size);
		}

		LDISKFS_I(inode)->i_inline_size = i_size <
					LDISKFS_MIN_INLINE_DATA_SIZE ?
					LDISKFS_MIN_INLINE_DATA_SIZE : i_size;
	}

out_error:
	up_write(&LDISKFS_I(inode)->i_data_sem);
out:
	brelse(is.iloc.bh);
	up_write(&LDISKFS_I(inode)->xattr_sem);
	kfree(value);
	if (inode->i_nlink)
		ldiskfs_orphan_del(handle, inode);

	inode->i_mtime = inode->i_ctime = ldiskfs_current_time(inode);
	ldiskfs_mark_inode_dirty(handle, inode);
	if (IS_SYNC(inode))
		ldiskfs_handle_sync(handle);

	ldiskfs_journal_stop(handle);
	return;
}

int ldiskfs_convert_inline_data(struct inode *inode)
{
	int error, needed_blocks;
	handle_t *handle;
	struct ldiskfs_iloc iloc;

	if (!ldiskfs_has_inline_data(inode)) {
		ldiskfs_clear_inode_state(inode, LDISKFS_STATE_MAY_INLINE_DATA);
		return 0;
	}

	needed_blocks = ldiskfs_writepage_trans_blocks(inode);

	iloc.bh = NULL;
	error = ldiskfs_get_inode_loc(inode, &iloc);
	if (error)
		return error;

	handle = ldiskfs_journal_start(inode, LDISKFS_HT_WRITE_PAGE, needed_blocks);
	if (IS_ERR(handle)) {
		error = PTR_ERR(handle);
		goto out_free;
	}

	down_write(&LDISKFS_I(inode)->xattr_sem);
	if (!ldiskfs_has_inline_data(inode)) {
		up_write(&LDISKFS_I(inode)->xattr_sem);
		goto out;
	}

	error = ldiskfs_convert_inline_data_nolock(handle, inode, &iloc);
	up_write(&LDISKFS_I(inode)->xattr_sem);
out:
	ldiskfs_journal_stop(handle);
out_free:
	brelse(iloc.bh);
	return error;
}
