/*
 * linux/fs/ldiskfs/ioctl.c
 *
 * Copyright (C) 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 */

#include <linux/fs.h>
#include <linux/jbd2.h>
#include <linux/capability.h>
#include <linux/time.h>
#include <linux/compat.h>
#include <linux/mount.h>
#include <linux/file.h>
#include <linux/uuid.h>
#include <asm/uaccess.h>
#include <linux/quotaops.h>
#include "ldiskfs_jbd2.h"
#include "ldiskfs.h"

#define MAX_32_NUM ((((unsigned long long) 1) << 32) - 1)

/**
 * Swap memory between @a and @b for @len bytes.
 *
 * @a:          pointer to first memory area
 * @b:          pointer to second memory area
 * @len:        number of bytes to swap
 *
 */
static void memswap(void *a, void *b, size_t len)
{
	unsigned char *ap, *bp;
	unsigned char tmp;

	ap = (unsigned char *)a;
	bp = (unsigned char *)b;
	while (len-- > 0) {
		tmp = *ap;
		*ap = *bp;
		*bp = tmp;
		ap++;
		bp++;
	}
}

/**
 * Swap i_data and associated attributes between @inode1 and @inode2.
 * This function is used for the primary swap between inode1 and inode2
 * and also to revert this primary swap in case of errors.
 *
 * Therefore you have to make sure, that calling this method twice
 * will revert all changes.
 *
 * @inode1:     pointer to first inode
 * @inode2:     pointer to second inode
 */
static void swap_inode_data(struct inode *inode1, struct inode *inode2)
{
	loff_t isize;
	struct ldiskfs_inode_info *ei1;
	struct ldiskfs_inode_info *ei2;

	ei1 = LDISKFS_I(inode1);
	ei2 = LDISKFS_I(inode2);

	memswap(&inode1->i_flags, &inode2->i_flags, sizeof(inode1->i_flags));
	memswap(&inode1->i_version, &inode2->i_version,
		  sizeof(inode1->i_version));
	memswap(&inode1->i_blocks, &inode2->i_blocks,
		  sizeof(inode1->i_blocks));
	memswap(&inode1->i_bytes, &inode2->i_bytes, sizeof(inode1->i_bytes));
	memswap(&inode1->i_atime, &inode2->i_atime, sizeof(inode1->i_atime));
	memswap(&inode1->i_mtime, &inode2->i_mtime, sizeof(inode1->i_mtime));

	memswap(ei1->i_data, ei2->i_data, sizeof(ei1->i_data));
	memswap(&ei1->i_flags, &ei2->i_flags, sizeof(ei1->i_flags));
	memswap(&ei1->i_disksize, &ei2->i_disksize, sizeof(ei1->i_disksize));
	ldiskfs_es_remove_extent(inode1, 0, EXT_MAX_BLOCKS);
	ldiskfs_es_remove_extent(inode2, 0, EXT_MAX_BLOCKS);
	ldiskfs_es_lru_del(inode1);
	ldiskfs_es_lru_del(inode2);

	isize = i_size_read(inode1);
	i_size_write(inode1, i_size_read(inode2));
	i_size_write(inode2, isize);
}

/**
 * Swap the information from the given @inode and the inode
 * LDISKFS_BOOT_LOADER_INO. It will basically swap i_data and all other
 * important fields of the inodes.
 *
 * @sb:         the super block of the filesystem
 * @inode:      the inode to swap with LDISKFS_BOOT_LOADER_INO
 *
 */
static long swap_inode_boot_loader(struct super_block *sb,
				struct inode *inode)
{
	handle_t *handle;
	int err;
	struct inode *inode_bl;
	struct ldiskfs_inode_info *ei_bl;
	struct ldiskfs_sb_info *sbi = LDISKFS_SB(sb);

	if (inode->i_nlink != 1 || !S_ISREG(inode->i_mode))
		return -EINVAL;

	if (!inode_owner_or_capable(inode) || !capable(CAP_SYS_ADMIN))
		return -EPERM;

	inode_bl = ldiskfs_iget(sb, LDISKFS_BOOT_LOADER_INO);
	if (IS_ERR(inode_bl))
		return PTR_ERR(inode_bl);
	ei_bl = LDISKFS_I(inode_bl);

	filemap_flush(inode->i_mapping);
	filemap_flush(inode_bl->i_mapping);

	/* Protect orig inodes against a truncate and make sure,
	 * that only 1 swap_inode_boot_loader is running. */
	lock_two_nondirectories(inode, inode_bl);

	truncate_inode_pages(&inode->i_data, 0);
	truncate_inode_pages(&inode_bl->i_data, 0);

	/* Wait for all existing dio workers */
	ldiskfs_inode_block_unlocked_dio(inode);
	ldiskfs_inode_block_unlocked_dio(inode_bl);
	inode_dio_wait(inode);
	inode_dio_wait(inode_bl);

	handle = ldiskfs_journal_start(inode_bl, LDISKFS_HT_MOVE_EXTENTS, 2);
	if (IS_ERR(handle)) {
		err = -EINVAL;
		goto journal_err_out;
	}

	/* Protect extent tree against block allocations via delalloc */
	ldiskfs_double_down_write_data_sem(inode, inode_bl);

	if (inode_bl->i_nlink == 0) {
		/* this inode has never been used as a BOOT_LOADER */
		set_nlink(inode_bl, 1);
		i_uid_write(inode_bl, 0);
		i_gid_write(inode_bl, 0);
		inode_bl->i_flags = 0;
		ei_bl->i_flags = 0;
		inode_bl->i_version = 1;
		i_size_write(inode_bl, 0);
		inode_bl->i_mode = S_IFREG;
		if (LDISKFS_HAS_INCOMPAT_FEATURE(sb,
					      LDISKFS_FEATURE_INCOMPAT_EXTENTS)) {
			ldiskfs_set_inode_flag(inode_bl, LDISKFS_INODE_EXTENTS);
			ldiskfs_ext_tree_init(handle, inode_bl);
		} else
			memset(ei_bl->i_data, 0, sizeof(ei_bl->i_data));
	}

	swap_inode_data(inode, inode_bl);

	inode->i_ctime = inode_bl->i_ctime = ldiskfs_current_time(inode);

	spin_lock(&sbi->s_next_gen_lock);
	inode->i_generation = sbi->s_next_generation++;
	inode_bl->i_generation = sbi->s_next_generation++;
	spin_unlock(&sbi->s_next_gen_lock);

	ldiskfs_discard_preallocations(inode);

	err = ldiskfs_mark_inode_dirty(handle, inode);
	if (err < 0) {
		ldiskfs_warning(inode->i_sb,
			"couldn't mark inode #%lu dirty (err %d)",
			inode->i_ino, err);
		/* Revert all changes: */
		swap_inode_data(inode, inode_bl);
	} else {
		err = ldiskfs_mark_inode_dirty(handle, inode_bl);
		if (err < 0) {
			ldiskfs_warning(inode_bl->i_sb,
				"couldn't mark inode #%lu dirty (err %d)",
				inode_bl->i_ino, err);
			/* Revert all changes: */
			swap_inode_data(inode, inode_bl);
			ldiskfs_mark_inode_dirty(handle, inode);
		}
	}
	ldiskfs_journal_stop(handle);
	ldiskfs_double_up_write_data_sem(inode, inode_bl);

journal_err_out:
	ldiskfs_inode_resume_unlocked_dio(inode);
	ldiskfs_inode_resume_unlocked_dio(inode_bl);
	unlock_two_nondirectories(inode, inode_bl);
	iput(inode_bl);
	return err;
}

static int ldiskfs_ioctl_setflags(struct inode *inode,
			       unsigned int flags)
{
	struct ldiskfs_inode_info *ei = LDISKFS_I(inode);
	handle_t *handle = NULL;
	int err = EPERM, migrate = 0;
	struct ldiskfs_iloc iloc;
	unsigned int oldflags, mask, i;
	unsigned int jflag;

	/* Is it quota file? Do not allow user to mess with it */
	if (IS_NOQUOTA(inode))
		goto flags_out;

	oldflags = ei->i_flags;

	/* The JOURNAL_DATA flag is modifiable only by root */
	jflag = flags & LDISKFS_JOURNAL_DATA_FL;

	/*
	 * The IMMUTABLE and APPEND_ONLY flags can only be changed by
	 * the relevant capability.
	 *
	 * This test looks nicer. Thanks to Pauline Middelink
	 */
	if ((flags ^ oldflags) & (LDISKFS_APPEND_FL | LDISKFS_IMMUTABLE_FL)) {
		if (!capable(CAP_LINUX_IMMUTABLE))
			goto flags_out;
	}

	/*
	 * The JOURNAL_DATA flag can only be changed by
	 * the relevant capability.
	 */
	if ((jflag ^ oldflags) & (LDISKFS_JOURNAL_DATA_FL)) {
		if (!capable(CAP_SYS_RESOURCE))
			goto flags_out;
	}
	if ((flags ^ oldflags) & LDISKFS_EXTENTS_FL)
		migrate = 1;

	if (flags & LDISKFS_EOFBLOCKS_FL) {
		/* we don't support adding EOFBLOCKS flag */
		if (!(oldflags & LDISKFS_EOFBLOCKS_FL)) {
			err = -EOPNOTSUPP;
			goto flags_out;
		}
	} else if (oldflags & LDISKFS_EOFBLOCKS_FL)
		ldiskfs_truncate(inode);

	handle = ldiskfs_journal_start(inode, LDISKFS_HT_INODE, 1);
	if (IS_ERR(handle)) {
		err = PTR_ERR(handle);
		goto flags_out;
	}
	if (IS_SYNC(inode))
		ldiskfs_handle_sync(handle);
	err = ldiskfs_reserve_inode_write(handle, inode, &iloc);
	if (err)
		goto flags_err;

	for (i = 0, mask = 1; i < 32; i++, mask <<= 1) {
		if (!(mask & LDISKFS_FL_USER_MODIFIABLE))
			continue;
		if (mask & flags)
			ldiskfs_set_inode_flag(inode, i);
		else
			ldiskfs_clear_inode_flag(inode, i);
	}

	ldiskfs_set_inode_flags(inode);
	inode->i_ctime = ldiskfs_current_time(inode);

	err = ldiskfs_mark_iloc_dirty(handle, inode, &iloc);
flags_err:
	ldiskfs_journal_stop(handle);
	if (err)
		goto flags_out;

	if ((jflag ^ oldflags) & (LDISKFS_JOURNAL_DATA_FL)) {
                        /*
                         * Changes to the journaling mode can cause unsafe
                         * changes to S_DAX if we are using the DAX mount option.
                         */
                        if (test_opt(inode->i_sb, DAX)) {
                                err = -EBUSY;
                                goto flags_out;
                        }
		err = ldiskfs_change_inode_journal_flag(inode, jflag);
	}
	if (err)
		goto flags_out;
	if (migrate) {
		if (flags & LDISKFS_EXTENTS_FL)
			err = ldiskfs_ext_migrate(inode);
		else
			err = ldiskfs_ind_migrate(inode);
	}

flags_out:
	return err;
}

#ifdef CONFIG_QUOTA
static int ldiskfs_ioctl_setproject(struct file *filp, __u32 projid)
{
	struct inode *inode = file_inode(filp);
	struct super_block *sb = inode->i_sb;
	struct ldiskfs_inode_info *ei = LDISKFS_I(inode);
	int err, rc;
	handle_t *handle;
	kprojid_t kprojid;
	struct ldiskfs_iloc iloc;
	struct ldiskfs_inode *raw_inode;
	struct dquot *transfer_to[LDISKFS_MAXQUOTAS] = { };

	if (!LDISKFS_HAS_RO_COMPAT_FEATURE(sb,
			LDISKFS_FEATURE_RO_COMPAT_PROJECT)) {
		BUG_ON(__kprojid_val(LDISKFS_I(inode)->i_projid)
		       != LDISKFS_DEF_PROJID);
		if (projid != LDISKFS_DEF_PROJID)
			return -EOPNOTSUPP;
		else
			return 0;
	}

	if (LDISKFS_INODE_SIZE(sb) <= LDISKFS_GOOD_OLD_INODE_SIZE)
		return -EOPNOTSUPP;

	kprojid = make_kprojid(&init_user_ns, (projid_t)projid);

	if (projid_eq(kprojid, LDISKFS_I(inode)->i_projid))
		return 0;

	err = mnt_want_write_file(filp);
	if (err)
		return err;

	err = -EPERM;
	mutex_lock(&inode->i_mutex);
	/* Is it quota file? Do not allow user to mess with it */
	if (IS_NOQUOTA(inode))
		goto out_unlock;

	err = ldiskfs_get_inode_loc(inode, &iloc);
	if (err)
		goto out_unlock;

	raw_inode = ldiskfs_raw_inode(&iloc);
	if (!LDISKFS_FITS_IN_INODE(raw_inode, ei, i_projid)) {
		err = -EOVERFLOW;
		brelse(iloc.bh);
		goto out_unlock;
	}
	brelse(iloc.bh);

	dquot_initialize(inode);

	handle = ldiskfs_journal_start(inode, LDISKFS_HT_QUOTA,
		LDISKFS_QUOTA_INIT_BLOCKS(sb) +
		LDISKFS_QUOTA_DEL_BLOCKS(sb) + 3);
	if (IS_ERR(handle)) {
		err = PTR_ERR(handle);
		goto out_unlock;
	}

	err = ldiskfs_reserve_inode_write(handle, inode, &iloc);
	if (err)
		goto out_stop;

	transfer_to[PRJQUOTA] = dqget(sb, make_kqid_projid(kprojid));
	if (transfer_to[PRJQUOTA]) {
		err = __dquot_transfer(inode, transfer_to);
		dqput(transfer_to[PRJQUOTA]);
		if (err)
			goto out_dirty;
	}

	LDISKFS_I(inode)->i_projid = kprojid;
	inode->i_ctime = ldiskfs_current_time(inode);
out_dirty:
	rc = ldiskfs_mark_iloc_dirty(handle, inode, &iloc);
	if (!err)
		err = rc;
out_stop:
	ldiskfs_journal_stop(handle);
out_unlock:
	mutex_unlock(&inode->i_mutex);
	mnt_drop_write_file(filp);
	return err;


}

#else
static int ldiskfs_ioctl_setproject(struct file *filp, __u32 projid)
{
	if (projid != LDISKFS_DEF_PROJID)
		return -EOPNOTSUPP;
	return 0;
}
#endif


/* Transfer internal flags to xflags */
static inline __u32 ldiskfs_iflags_to_xflags(unsigned long iflags)
{
	__u32 xflags = 0;

	if (iflags & LDISKFS_SYNC_FL)
		xflags |= FS_XFLAG_SYNC;
	if (iflags & LDISKFS_IMMUTABLE_FL)
		xflags |= FS_XFLAG_IMMUTABLE;
	if (iflags & LDISKFS_APPEND_FL)
		xflags |= FS_XFLAG_APPEND;
	if (iflags & LDISKFS_NODUMP_FL)
		xflags |= FS_XFLAG_NODUMP;
	if (iflags & LDISKFS_NOATIME_FL)
		xflags |= FS_XFLAG_NOATIME;
	if (iflags & LDISKFS_PROJINHERIT_FL)
		xflags |= FS_XFLAG_PROJINHERIT;
	return xflags;
}

/* Transfer xflags flags to internal */
static inline unsigned long ldiskfs_xflags_to_iflags(__u32 xflags)
{
	unsigned long iflags = 0;

	if (xflags & FS_XFLAG_SYNC)
		iflags |= LDISKFS_SYNC_FL;
	if (xflags & FS_XFLAG_IMMUTABLE)
		iflags |= LDISKFS_IMMUTABLE_FL;
	if (xflags & FS_XFLAG_APPEND)
		iflags |= LDISKFS_APPEND_FL;
	if (xflags & FS_XFLAG_NODUMP)
		iflags |= LDISKFS_NODUMP_FL;
	if (xflags & FS_XFLAG_NOATIME)
		iflags |= LDISKFS_NOATIME_FL;
	if (xflags & FS_XFLAG_PROJINHERIT)
		iflags |= LDISKFS_PROJINHERIT_FL;

	return iflags;
}

long ldiskfs_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct inode *inode = file_inode(filp);
	struct super_block *sb = inode->i_sb;
	struct ldiskfs_inode_info *ei = LDISKFS_I(inode);
	unsigned int flags;

	ldiskfs_debug("cmd = %u, arg = %lu\n", cmd, arg);

	switch (cmd) {
	case LDISKFS_IOC_GETFLAGS:
		ldiskfs_get_inode_flags(ei);
		flags = ei->i_flags & LDISKFS_FL_USER_VISIBLE;
		return put_user(flags, (int __user *) arg);
	case LDISKFS_IOC_SETFLAGS: {
		int err;

		if (!inode_owner_or_capable(inode))
			return -EACCES;

		if (get_user(flags, (int __user *) arg))
			return -EFAULT;

		err = mnt_want_write_file(filp);
		if (err)
			return err;

		flags = ldiskfs_mask_flags(inode->i_mode, flags);

		mutex_lock(&inode->i_mutex);
		err = ldiskfs_ioctl_setflags(inode, flags);
		mutex_unlock(&inode->i_mutex);
		mnt_drop_write_file(filp);
		return err;
	}
	case LDISKFS_IOC_GETVERSION:
	case LDISKFS_IOC_GETVERSION_OLD:
		return put_user(inode->i_generation, (int __user *) arg);
	case LDISKFS_IOC_SETVERSION:
	case LDISKFS_IOC_SETVERSION_OLD: {
		handle_t *handle;
		struct ldiskfs_iloc iloc;
		__u32 generation;
		int err;

		if (!inode_owner_or_capable(inode))
			return -EPERM;

		if (ldiskfs_has_metadata_csum(inode->i_sb)) {
			ldiskfs_warning(sb, "Setting inode version is not "
				     "supported with metadata_csum enabled.");
			return -ENOTTY;
		}

		err = mnt_want_write_file(filp);
		if (err)
			return err;
		if (get_user(generation, (int __user *) arg)) {
			err = -EFAULT;
			goto setversion_out;
		}

		mutex_lock(&inode->i_mutex);
		handle = ldiskfs_journal_start(inode, LDISKFS_HT_INODE, 1);
		if (IS_ERR(handle)) {
			err = PTR_ERR(handle);
			goto unlock_out;
		}
		err = ldiskfs_reserve_inode_write(handle, inode, &iloc);
		if (err == 0) {
			inode->i_ctime = ldiskfs_current_time(inode);
			inode->i_generation = generation;
			err = ldiskfs_mark_iloc_dirty(handle, inode, &iloc);
		}
		ldiskfs_journal_stop(handle);

unlock_out:
		mutex_unlock(&inode->i_mutex);
setversion_out:
		mnt_drop_write_file(filp);
		return err;
	}
	case LDISKFS_IOC_GROUP_EXTEND: {
		ldiskfs_fsblk_t n_blocks_count;
		int err, err2=0;

		err = ldiskfs_resize_begin(sb);
		if (err)
			return err;

		if (get_user(n_blocks_count, (__u32 __user *)arg)) {
			err = -EFAULT;
			goto group_extend_out;
		}

		if (LDISKFS_HAS_RO_COMPAT_FEATURE(sb,
			       LDISKFS_FEATURE_RO_COMPAT_BIGALLOC)) {
			ldiskfs_msg(sb, KERN_ERR,
				 "Online resizing not supported with bigalloc");
			err = -EOPNOTSUPP;
			goto group_extend_out;
		}

		err = mnt_want_write_file(filp);
		if (err)
			goto group_extend_out;

		err = ldiskfs_group_extend(sb, LDISKFS_SB(sb)->s_es, n_blocks_count);
		if (LDISKFS_SB(sb)->s_journal) {
			jbd2_journal_lock_updates(LDISKFS_SB(sb)->s_journal);
			err2 = jbd2_journal_flush(LDISKFS_SB(sb)->s_journal);
			jbd2_journal_unlock_updates(LDISKFS_SB(sb)->s_journal);
		}
		if (err == 0)
			err = err2;
		mnt_drop_write_file(filp);
group_extend_out:
		ldiskfs_resize_end(sb);
		return err;
	}

	case LDISKFS_IOC_MOVE_EXT: {
		struct move_extent me;
		struct fd donor;
		int err;

		if (!(filp->f_mode & FMODE_READ) ||
		    !(filp->f_mode & FMODE_WRITE))
			return -EBADF;

		if (copy_from_user(&me,
			(struct move_extent __user *)arg, sizeof(me)))
			return -EFAULT;
		me.moved_len = 0;

		donor = fdget(me.donor_fd);
		if (!donor.file)
			return -EBADF;

		if (!(donor.file->f_mode & FMODE_WRITE)) {
			err = -EBADF;
			goto mext_out;
		}

		if (LDISKFS_HAS_RO_COMPAT_FEATURE(sb,
			       LDISKFS_FEATURE_RO_COMPAT_BIGALLOC)) {
			ldiskfs_msg(sb, KERN_ERR,
				 "Online defrag not supported with bigalloc");
			err = -EOPNOTSUPP;
			goto mext_out;
		} else if (IS_DAX(inode)) {
			ldiskfs_msg(sb, KERN_ERR,
				 "Online defrag not supported with DAX");
			err = -EOPNOTSUPP;
			goto mext_out;
		}

		err = mnt_want_write_file(filp);
		if (err)
			goto mext_out;

		err = ldiskfs_move_extents(filp, donor.file, me.orig_start,
					me.donor_start, me.len, &me.moved_len);
		mnt_drop_write_file(filp);

		if (copy_to_user((struct move_extent __user *)arg,
				 &me, sizeof(me)))
			err = -EFAULT;
mext_out:
		fdput(donor);
		return err;
	}

	case LDISKFS_IOC_GROUP_ADD: {
		struct ldiskfs_new_group_data input;
		int err, err2=0;

		err = ldiskfs_resize_begin(sb);
		if (err)
			return err;

		if (copy_from_user(&input, (struct ldiskfs_new_group_input __user *)arg,
				sizeof(input))) {
			err = -EFAULT;
			goto group_add_out;
		}

		if (LDISKFS_HAS_RO_COMPAT_FEATURE(sb,
			       LDISKFS_FEATURE_RO_COMPAT_BIGALLOC)) {
			ldiskfs_msg(sb, KERN_ERR,
				 "Online resizing not supported with bigalloc");
			err = -EOPNOTSUPP;
			goto group_add_out;
		}

		err = mnt_want_write_file(filp);
		if (err)
			goto group_add_out;

		err = ldiskfs_group_add(sb, &input);
		if (LDISKFS_SB(sb)->s_journal) {
			jbd2_journal_lock_updates(LDISKFS_SB(sb)->s_journal);
			err2 = jbd2_journal_flush(LDISKFS_SB(sb)->s_journal);
			jbd2_journal_unlock_updates(LDISKFS_SB(sb)->s_journal);
		}
		if (err == 0)
			err = err2;
		mnt_drop_write_file(filp);
		if (!err && ldiskfs_has_group_desc_csum(sb) &&
		    test_opt(sb, INIT_INODE_TABLE))
			err = ldiskfs_register_li_request(sb, input.group);
group_add_out:
		ldiskfs_resize_end(sb);
		return err;
	}

	case LDISKFS_IOC_MIGRATE:
	{
		int err;
		if (!inode_owner_or_capable(inode))
			return -EACCES;

		err = mnt_want_write_file(filp);
		if (err)
			return err;
		/*
		 * inode_mutex prevent write and truncate on the file.
		 * Read still goes through. We take i_data_sem in
		 * ldiskfs_ext_swap_inode_data before we switch the
		 * inode format to prevent read.
		 */
		mutex_lock(&(inode->i_mutex));
		err = ldiskfs_ext_migrate(inode);
		mutex_unlock(&(inode->i_mutex));
		mnt_drop_write_file(filp);
		return err;
	}

	case LDISKFS_IOC_ALLOC_DA_BLKS:
	{
		int err;
		if (!inode_owner_or_capable(inode))
			return -EACCES;

		err = mnt_want_write_file(filp);
		if (err)
			return err;
		err = ldiskfs_alloc_da_blocks(inode);
		mnt_drop_write_file(filp);
		return err;
	}

	case LDISKFS_IOC_SWAP_BOOT:
	{
		int err;
		if (!(filp->f_mode & FMODE_WRITE))
			return -EBADF;
		err = mnt_want_write_file(filp);
		if (err)
			return err;
		err = swap_inode_boot_loader(sb, inode);
		mnt_drop_write_file(filp);
		return err;
	}

	case LDISKFS_IOC_RESIZE_FS: {
		ldiskfs_fsblk_t n_blocks_count;
		int err = 0, err2 = 0;
		ldiskfs_group_t o_group = LDISKFS_SB(sb)->s_groups_count;

		if (LDISKFS_HAS_RO_COMPAT_FEATURE(sb,
			       LDISKFS_FEATURE_RO_COMPAT_BIGALLOC)) {
			ldiskfs_msg(sb, KERN_ERR,
				 "Online resizing not (yet) supported with bigalloc");
			return -EOPNOTSUPP;
		}

		if (copy_from_user(&n_blocks_count, (__u64 __user *)arg,
				   sizeof(__u64))) {
			return -EFAULT;
		}

		err = ldiskfs_resize_begin(sb);
		if (err)
			return err;

		err = mnt_want_write_file(filp);
		if (err)
			goto resizefs_out;

		err = ldiskfs_resize_fs(sb, n_blocks_count);
		if (LDISKFS_SB(sb)->s_journal) {
			jbd2_journal_lock_updates(LDISKFS_SB(sb)->s_journal);
			err2 = jbd2_journal_flush(LDISKFS_SB(sb)->s_journal);
			jbd2_journal_unlock_updates(LDISKFS_SB(sb)->s_journal);
		}
		if (err == 0)
			err = err2;
		mnt_drop_write_file(filp);
		if (!err && (o_group > LDISKFS_SB(sb)->s_groups_count) &&
		    ldiskfs_has_group_desc_csum(sb) &&
		    test_opt(sb, INIT_INODE_TABLE))
			err = ldiskfs_register_li_request(sb, o_group);

resizefs_out:
		ldiskfs_resize_end(sb);
		return err;
	}

	case FITRIM:
	{
		struct request_queue *q = bdev_get_queue(sb->s_bdev);
		struct fstrim_range range;
		int ret = 0;

		if (!capable(CAP_SYS_ADMIN))
			return -EPERM;

		if (!blk_queue_discard(q))
			return -EOPNOTSUPP;

		if (copy_from_user(&range, (struct fstrim_range __user *)arg,
		    sizeof(range)))
			return -EFAULT;

		range.minlen = max((unsigned int)range.minlen,
				   q->limits.discard_granularity);
		ret = ldiskfs_trim_fs(sb, &range);
		if (ret < 0)
			return ret;

		if (copy_to_user((struct fstrim_range __user *)arg, &range,
		    sizeof(range)))
			return -EFAULT;

		return 0;
	}
	case LDISKFS_IOC_PRECACHE_EXTENTS:
		return ldiskfs_ext_precache(inode);
	case LDISKFS_IOC_FSGETXATTR:
	{
		struct fsxattr fa;
		unsigned int flags;

		memset(&fa, 0, sizeof(struct fsxattr));
		ldiskfs_get_inode_flags(ei);
		flags = ei->i_flags & LDISKFS_FL_USER_VISIBLE;
		fa.fsx_xflags = ldiskfs_iflags_to_xflags(flags);

		if (LDISKFS_HAS_RO_COMPAT_FEATURE(inode->i_sb,
				LDISKFS_FEATURE_RO_COMPAT_PROJECT)) {
			fa.fsx_projid = (__u32)from_kprojid(&init_user_ns,
				LDISKFS_I(inode)->i_projid);
		}

		if (copy_to_user((struct fsxattr __user *)arg,
				 &fa, sizeof(fa)))
			return -EFAULT;
		return 0;
	}
	case LDISKFS_IOC_FSSETXATTR:
	{
		struct fsxattr fa;
		int err;

		if (copy_from_user(&fa, (struct fsxattr __user *)arg,
				   sizeof(fa)))
			return -EFAULT;

		/* Make sure caller has proper permission */
		if (!inode_owner_or_capable(inode))
			return -EACCES;

		err = mnt_want_write_file(filp);
		if (err)
			return err;

		flags = ldiskfs_xflags_to_iflags(fa.fsx_xflags);
		flags = ldiskfs_mask_flags(inode->i_mode, flags);

		mutex_lock(&inode->i_mutex);
		flags = (ei->i_flags & ~LDISKFS_FL_XFLAG_VISIBLE) |
			 (flags & LDISKFS_FL_XFLAG_VISIBLE);
		err = ldiskfs_ioctl_setflags(inode, flags);
		mutex_unlock(&inode->i_mutex);
		mnt_drop_write_file(filp);
		if (err)
			return err;

		err = ldiskfs_ioctl_setproject(filp, fa.fsx_projid);
		if (err)
			return err;

		return 0;
	}

	default:
		return -ENOTTY;
	}
}

#ifdef CONFIG_COMPAT
long ldiskfs_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	/* These are just misnamed, they actually get/put from/to user an int */
	switch (cmd) {
	case LDISKFS_IOC32_GETFLAGS:
		cmd = LDISKFS_IOC_GETFLAGS;
		break;
	case LDISKFS_IOC32_SETFLAGS:
		cmd = LDISKFS_IOC_SETFLAGS;
		break;
	case LDISKFS_IOC32_GETVERSION:
		cmd = LDISKFS_IOC_GETVERSION;
		break;
	case LDISKFS_IOC32_SETVERSION:
		cmd = LDISKFS_IOC_SETVERSION;
		break;
	case LDISKFS_IOC32_GROUP_EXTEND:
		cmd = LDISKFS_IOC_GROUP_EXTEND;
		break;
	case LDISKFS_IOC32_GETVERSION_OLD:
		cmd = LDISKFS_IOC_GETVERSION_OLD;
		break;
	case LDISKFS_IOC32_SETVERSION_OLD:
		cmd = LDISKFS_IOC_SETVERSION_OLD;
		break;
	case LDISKFS_IOC32_GETRSVSZ:
		cmd = LDISKFS_IOC_GETRSVSZ;
		break;
	case LDISKFS_IOC32_SETRSVSZ:
		cmd = LDISKFS_IOC_SETRSVSZ;
		break;
	case LDISKFS_IOC32_GROUP_ADD: {
		struct compat_ldiskfs_new_group_input __user *uinput;
		struct ldiskfs_new_group_input input;
		mm_segment_t old_fs;
		int err;

		uinput = compat_ptr(arg);
		err = get_user(input.group, &uinput->group);
		err |= get_user(input.block_bitmap, &uinput->block_bitmap);
		err |= get_user(input.inode_bitmap, &uinput->inode_bitmap);
		err |= get_user(input.inode_table, &uinput->inode_table);
		err |= get_user(input.blocks_count, &uinput->blocks_count);
		err |= get_user(input.reserved_blocks,
				&uinput->reserved_blocks);
		if (err)
			return -EFAULT;
		old_fs = get_fs();
		set_fs(KERNEL_DS);
		err = ldiskfs_ioctl(file, LDISKFS_IOC_GROUP_ADD,
				 (unsigned long) &input);
		set_fs(old_fs);
		return err;
	}
	case LDISKFS_IOC_MOVE_EXT:
	case FITRIM:
	case LDISKFS_IOC_RESIZE_FS:
	case LDISKFS_IOC_PRECACHE_EXTENTS:
		break;
	default:
		return -ENOIOCTLCMD;
	}
	return ldiskfs_ioctl(file, cmd, (unsigned long) compat_ptr(arg));
}
#endif
